function symbols = LoRa_encoder(payload, SF, CR, CRC_on, DE)
%LORA_ENCODER encodes a payload to LoRa symbols
%   the function creates LoRa symbols for a vector of decimal payload with
%   length 0 ... 255 with spreading factor SF (1 ...7) and code rate CR. 
%   CR = 1 ... code rate 4/5
%   CR = 2 ... code rate 4/6
%   CR = 3 ... code rate 4/7
%   CR = 4 ... code rate 4/8
%   CRC_on = 1 indicates the presence of a payload CRC. DE = 1 indicates the activated
%   low data rate mode

% Authors: Holger Arthaber, Harald Eigner
%          TU Wien (EMCE), www.emce.tuwien.ac.at

% This work is licensed under a Creative Commons Attribution 4.0 International 
% (CC BY 4.0), https://creativecommons.org/licenses/by/4.0/
% You are free to:
%  - Share — copy and redistribute the material in any medium or format
%  - Adapt — remix, transform, and build upon the material for any purpose, 
%    even commercially. 
% Under the following terms:
%  - Attribution — You must give appropriate credit, provide a link to the license, 
%    and indicate if changes were made. You may do so in any reasonable manner, but not 
%    in any way that suggests the licensor endorses you or your use. 

if ~isvector(payload) || any(payload < 0) || any(payload > 255) || any(payload ~= floor(payload))
    error('PAYLOAD must be a vector of integers in the uint8-rang 0...255');
end
if ~isvector(SF) || any(SF < 7) || any(SF > 12) || any(SF ~= floor(SF))
    error('SF must be a integer the uint8-rang 7...12');
end
if ~isvector(CR) || any(CR < 1) || any(CR > 4) || any(CR ~= floor(CR))
    error('CR must be a integer the uint8-rang 7...12');
end



payload_val = true;                     
payload_hex = dec2hex(payload, 2);   %transform payload to hex format
payload_len = size(payload_hex, 1);  %define payload length

%calculation of symbol length
symbols_len_cal = 8 + max(ceil((8 * payload_len - 4 * SF + 28 + CRC_on * 16) / (4 * (SF - 2 * DE)))*(CR + 4), 0);  


%% Payload Whitening

payload_whitened_dec = compute_whitening(payload)';     %payload whitened dec
payload_whitened_hex = dec2hex(payload_whitened_dec);   %payload whitened hex

%% Calculate/Add CRC

if CRC_on == 1
    CRC_dec = payload_CRC(payload);   %calculate CRC dec
    CRC_hex = flipud(reshape(dec2hex(CRC_dec, 4)', 2, 2)');    %CRC hex
    payload_whitened_hex = [payload_whitened_hex; CRC_hex]; %add CRC to payload whitened hex
    payload_whitened_dec = hex2dec(payload_whitened_hex);   %payload whitened dec
end

payload_whitened_bin = zeros(payload_len + 2 * CRC_on, 8);
for idx = 1 : payload_len + 2 * CRC_on
    payload_whitened_bin(idx, :) = dec2binvec(payload_whitened_dec(idx)', 8);   %payload whitened binary
end

payload_whitened_bin = reshape(payload_whitened_bin', 4, (payload_len + 2 * CRC_on) * 2)';   %transform from byte to 4 bit nibble


%% Header Encoding

header_bin = header_encode(payload_len, CRC_on, CR + 4);   %create header bits

if SF > 7     %fill remaining space in first 8 symbols with payload   
    if size(payload_whitened_bin, 1) <= SF - 7
        header_bin=[header_bin; payload_whitened_bin; zeros(SF - 7 - size(payload_whitened_bin, 1) , 4)];
        payload_val=false;
    else
        header_bin=[header_bin; payload_whitened_bin(1 : SF - 7, :)];
        payload_whitened_bin = payload_whitened_bin(SF - 6 : end, :);
    end
    
end

header_encoded = encoder(header_bin, 8);      %Encoding, reduced
header_interleaved = interleave(header_encoded);         %Interleaving

header_gray_bin = zeros(8, SF - 2);
for i = 1 : 8
    header_gray_bin(i, :)=gray2bin(header_interleaved(i, :));     %Gray-Encoding
end

header_sym = bi2de(header_gray_bin) * 4;    %header symbols (non-reduced symbol alphabet SF-2)


%% Payload Encoding

if payload_val == true
    blocks = (symbols_len_cal - 8) / (CR + 4);    %Calculate ammount of codeword blocks
    
    payload_whitened_bin=[payload_whitened_bin; false(blocks * ( SF - 2 * DE) - size(payload_whitened_bin, 1), 4)]; %fill rest of payload with zeros to achive the calculated amount of symbols
    
    payload_encoded = zeros(blocks * SF, CR + 4);
    payload_interleaved = zeros(blocks * (CR + 4), SF);
    for idx = 1 : blocks
        SF_ = (1 : SF - 2 * DE) + (idx - 1) * (SF - 2 * DE);
        CR_ = (1 : CR + 4) + (idx - 1) * (CR + 4);
        
        payload_encoded(SF_, :) = encoder(payload_whitened_bin(SF_, :),CR + 4);  %Encoding
        payload_interleaved(CR_, 1:(SF - 2 * DE)) = interleave(payload_encoded(SF_, 1:(CR + 4)));     %interleaving
    end
    
    payload_gray_bin = zeros(blocks * (CR + 4), SF);
    for i = 1 : size(payload_interleaved, 1)
        payload_gray_bin(i, :) = gray2bin(payload_interleaved(i, :));   %Gray-Encoding
    end
    
    payload_sym=bi2de(payload_gray_bin);    %payload symbols
    
    if DE == 1
        payload_sym = payload_sym * 4;      %transform payload symbols to non-reduced symbol alphabet
    end
    
    symbols = ([header_sym; payload_sym])';
else
    symbols = header_sym';
end

end





