function [decoded] = decoder(B, CR)
%ENCODE Data decoding
%   This function decodes the input data from matrix B with code rate:
%   CR = 1 ... code rate 4/5
%   CR = 2 ... code rate 4/6
%   CR = 3 ... code rate 4/7
%   CR = 4 ... code rate 4/8
%   The check matrix for the according code rate is calculated and the
%   validity of the codewords in the matrix B is checked. If the codewords
%   are valid, the data bits are returned.

% Authors: Holger Arthaber, Harald Eigner
%          TU Wien (EMCE), www.emce.tuwien.ac.at

% This work is licensed under a Creative Commons Attribution 4.0 International 
% (CC BY 4.0), https://creativecommons.org/licenses/by/4.0/
% You are free to:
%  - Share — copy and redistribute the material in any medium or format
%  - Adapt — remix, transform, and build upon the material for any purpose, 
%    even commercially. 
% Under the following terms:
%  - Attribution — You must give appropriate credit, provide a link to the license, 
%    and indicate if changes were made. You may do so in any reasonable manner, but not 
%    in any way that suggests the licensor endorses you or your use. 


if ~ismatrix(B) || any(B < 0, 'all') || any(B > 1, 'all') || any(B ~= floor(B), 'all')
    error('Invalid Codeblock');
end
if ~isscalar(CR) || CR < 5 || CR > 8 || CR ~= floor(CR)
    error('Invalid Code Rate');
end

len = size(B, 1);
N = CR;
k = 4;
decoded = zeros(len, k);
P = zeros(k, N - k);
I = logical(eye(k));
check = zeros(len, N);

%Create check matrix H
if CR > 5
    for idx = 1:4
        P(idx, 1) =  xor(xor(I(idx, 1), I(idx, 2)), I(idx, 3));
        P(idx, 2) =  xor(xor(I(idx, 2), I(idx, 3)), I(idx, 4));
        P(idx, 3) =  xor(xor(I(idx, 1), I(idx, 2)), I(idx, 4));
        P(idx, 4) =  xor(xor(I(idx, 1), I(idx, 3)), I(idx, 4));
        H = [P(1 : 4, 1 : (CR - 4)).' logical(eye(N - k))];    %check matrix for CR = 1 ... 3
    end
elseif CR == 5
    for idx = 1:4
        P(idx, 1) =  xor(xor(I(idx, 1), I(idx, 2)), xor(I(idx, 3), I(idx, 4)));
        H=[P.' logical(eye(N - k))];    %check matrix for CR = 4
    end
end

for i = 1:len
    for j = 1:N - k
        check(i, j)=mod(sum(B(i, :).*H(j, :)), 2);  %multiplication of codeword with checkmatrix c*H'=0
    end
end

%Check validity of header
if sum(check == 0)
    decoded = B(1:len, 1:k);
else
    error('Invalid Header');
end
end